#######################################################################
###### Plugin Software 'BatchLR' was Designed and Developed by
###### BDLS Informatics, Product Innovation.
###### Authored by Joshua Luthy, and Ian Taylor
###### This template was last updated in August 2019
###### The underlying R package 'Batchelor', and Mutual Nearest
###### Neighbors algorithm on which that package is based, was
###### developed by: Aaron Lun, Cancer Research UK Cambridge Institute
#######################################################################


#######################################################################
###### Start package import
#######################################################################

# Load libraries
## Instructions to Install dplyr:
# install.packages("dplyr")

library(dplyr)

## Instructions to Install batchelor:
# if (!requireNamespace("BiocManager", quietly = TRUE))
 # install.packages("BiocManager")
 # BiocManager::install("batchelor")

library(batchelor)
library(scater)
library(SingleCellExperiment)

## Instructions to install scran:
# if (!requireNamespace("BiocManager", quietly = TRUE))
#   install.packages("BiocManager")
# BiocManager::install("scran")

library(scran)

setwd("SG_OUTPUT_DIRECTORY_PATH")
#######################################################################
###### Start variable import
#######################################################################
knn <- KNN

# Test knn; comment this in to test:
# knn <- 16

# Unused Var:
# dist <- DIST

#######################################################################
###### Start the File Import
#######################################################################
# Read CSV from SeqGeq

importedCSVfile <- read.table(file= "SG_SAMPLE_DATA_PATH", sep = ",", row.names = 1, header = TRUE, check.names = FALSE)
importedCSVfile = importedCSVfile[, colSums(importedCSVfile != 0) > 0]

# Cell and Feature Annotations:
cellAnnotationsFrame <- data.frame(colnames(importedCSVfile))
featureAnnotationsFrame <- data.frame(rownames(importedCSVfile))
colnames(cellAnnotationsFrame) <- paste("CellID")
# Some data has numbers as cellIDs only, append a string to cellIds, remove on export.
colnames(importedCSVfile) <- paste("AppID", colnames(importedCSVfile),sep="_")
batchFrame <- importedCSVfile["SampleID",]

# new <- importedCSVfile[rownames(-importedCSVfile),]
###############################################################
#  Following SC3 package manual for data input (below)   #
###############################################################

writeLines(" Start Writing Batchelor SCE object ")
# create a SingleCellExperiment object
sce <- SingleCellExperiment(list(counts=as.matrix(importedCSVfile),
                                  logcounts = log2(as.matrix(importedCSVfile) + 1)
),
colData=cellAnnotationsFrame,
rowData=featureAnnotationsFrame)
# Define batchIDs
colData(sce)$batchID <- batchFrame

#################################################################
#   Performing Batch Correction on two batches
#################################################################
# Performing fastMNN batch correction on the two imported SeqGeq populations.
# directly for downstream analyses such as clustering and visualization.

mnn.out <- batchelor::fastMNN(sce, batch = batchFrame, k=knn)
# TODO: Give a Nice error message if Waning: "tied distances detected in nearest-neighbor calculation" is thrown.

# Display string of corrected values

str(reducedDim(mnn.out, "corrected"))

#################################################################
##### Viewing the Batch Corrected Pops
#################################################################
# Each column of mnn.out corresponds to a cell in one of the batches, 
# while each row corresponds to an input gene in chosen. The corrected 
# matrix in the reducedDims slot contains the low-dimensional corrected coordinates for all cells.
batches <- list(mnn.out)
head(batches)
dim(reducedDim(mnn.out, "corrected"))
reducedDims(mnn.out)

# Using an (run length encoding) Rle for pretty-printing of batch IDs
# (as all cells from the same batch are consecutive).

Rle(mnn.out$batch) 

#################################################################
##### Effect of correction with Diagnostics
#################################################################
# One useful diagnostic is the proportion of variance within each batch 
# that is lost during MNN correction. Specifically, this refers to the 
# within-batch variance that is removed during orthogonalization with 
# respect to the average correction vector at each merge step. This is 
# returned via the lost.var field in the metadata of mnn.out, which contains 
# a matrix of the variance lost in each batch (column) at each merge step (row).

Lost_Variance <- metadata(mnn.out)$merge.info$lost.var

writeLines(" Difference in variance lost between batches:  ")
Lost_Variance
writeLines(" Large values, and major differences between batches indicates a loss of true biological signal. ")
writeLines(" Differences between the batch effect and biological signal are assumed to be orthagonal.(Haghverdi et al. 2018) ")

## EG of "good values", low/similar losses:
##             [,1]       [,2]
## [1,] 0.02094553 0.03466084

# Large proportions of lost variance suggest that correction is removing genuine 
# biological heterogeneity. This would occur due to violations of the assumption of 
# orthogonality between the batch effect and the biological subspace (Haghverdi et al. 2018). 

# # TODO What is considered a small variance here? TODO
# TODO Print these values out for each batch, in a sensible way TODO

#################################################################
#   Plotting two Batch Corrected Populations  #
#################################################################
# plotReducedDim(mnn.out, colour_by="batch", use_dimred = "corrected")

#################################################################
#   Create a Header for the first 5 PC  #
#################################################################

MNNpcHeader <- data.frame()
for(i in 1:25){
  de_i <- list(paste("MNNpc","SG_RUNID",i))
  MNNpcHeader <- rbind(MNNpcHeader,de_i,stringsAsFactors=FALSE,head(NULL))
}

MNNpcHeader_clean <- as.data.frame(MNNpcHeader, row.names=NULL)

write.table(MNNpcHeader_clean,paste0("MNN-PCA_","SG_RUNID",".csv"), quote=FALSE, row.names=FALSE, col.names=FALSE)

#################################################################
#   Exporting the corrected values to a .csv  #
#################################################################

pcMNNout <- as.data.frame(reducedDim(mnn.out, "corrected"))
MNNpcOut <- pcMNNout[,1:25]
colnames(MNNpcOut) <- paste(MNNpcHeader$X.MNNpc.SG_RUNID.1)
cAF3 <- data.frame(cellAnnotationsFrame)
rownames(MNNpcOut) <- paste(cAF3$CellID)
write.csv(MNNpcOut, paste0("BatchLR_MNN-PCA_","SG_RUNID",".csv"))
